@@ -1,19 +1,21 @@
 """Provide strategies for given endpoint(s) definition."""
 import asyncio
+import inspect
 import re
 from base64 import b64encode
 from functools import partial
-from typing import Any, Callable, Dict, Optional, Union
+from typing import Any, Callable, Dict, Optional, Tuple, Union
 from urllib.parse import quote_plus
 
 import hypothesis
 import hypothesis.strategies as st
 from hypothesis_jsonschema import from_schema
+from requests.auth import _basic_auth_str
 
 from . import utils
 from ._compat import handle_warnings
 from .exceptions import InvalidSchema
-from .hooks import get_hook
+from .hooks import GLOBAL_HOOK_DISPATCHER, HookContext, HookDispatcher
 from .models import Case, Endpoint
 from .types import Hook
 
@@ -25,7 +27,8 @@ def create_test(
     endpoint: Endpoint, test: Callable, settings: Optional[hypothesis.settings] = None, seed: Optional[int] = None
 ) -> Callable:
     """Create a Hypothesis test."""
-    strategy = endpoint.as_strategy()
+    hook_dispatcher = getattr(test, "_schemathesis_hooks", None)
+    strategy = endpoint.as_strategy(hooks=hook_dispatcher)
     wrapped_test = hypothesis.given(case=strategy)(test)
     if seed is not None:
         wrapped_test = hypothesis.seed(seed)(wrapped_test)
@@ -119,7 +122,7 @@ def is_valid_query(query: Dict[str, Any]) -> bool:
     return True
 
 
-def get_case_strategy(endpoint: Endpoint) -> st.SearchStrategy:
+def get_case_strategy(endpoint: Endpoint, hooks: Optional[HookDispatcher] = None) -> st.SearchStrategy:
     """Create a strategy for a complete test case.
 
     Path & endpoint are static, the others are JSON schemas.
@@ -142,7 +145,7 @@ def get_case_strategy(endpoint: Endpoint) -> st.SearchStrategy:
                     strategies[parameter] = from_schema(value)  # type: ignore
             else:
                 static_kwargs[parameter] = None
-        return _get_case_strategy(endpoint, static_kwargs, strategies)
+        return _get_case_strategy(endpoint, static_kwargs, strategies, hooks)
     except AssertionError:
         raise InvalidSchema("Invalid schema for this endpoint")
 
@@ -158,11 +161,7 @@ def filter_path_parameters(parameters: Dict[str, Any]) -> bool:
     Because of it this case doesn't bring much value and might lead to false positives results of Schemathesis runs.
     """
 
-    path_parameter_blacklist = (
-        ".",
-        SLASH,
-        "",
-    )
+    path_parameter_blacklist = (".", SLASH, "")
 
     return not any(
         (value in path_parameter_blacklist or isinstance(value, str) and SLASH in value)
@@ -175,7 +174,10 @@ def quote_all(parameters: Dict[str, Any]) -> Dict[str, Any]:
 
 
 def _get_case_strategy(
-    endpoint: Endpoint, extra_static_parameters: Dict[str, Any], strategies: Dict[str, st.SearchStrategy]
+    endpoint: Endpoint,
+    extra_static_parameters: Dict[str, Any],
+    strategies: Dict[str, st.SearchStrategy],
+    hook_dispatcher: Optional[HookDispatcher] = None,
 ) -> st.SearchStrategy:
     static_parameters: Dict[str, Any] = {"endpoint": endpoint, **extra_static_parameters}
     if endpoint.schema.validate_schema and endpoint.method == "GET":
@@ -183,16 +185,30 @@ def _get_case_strategy(
             raise InvalidSchema("Body parameters are defined for GET request.")
         static_parameters["body"] = None
         strategies.pop("body", None)
-    _apply_hooks(strategies, get_hook)
-    _apply_hooks(strategies, endpoint.schema.get_hook)
+    context = HookContext(endpoint)
+    _apply_hooks(strategies, GLOBAL_HOOK_DISPATCHER, context)
+    _apply_hooks(strategies, endpoint.schema.hooks, context)
+    if hook_dispatcher is not None:
+        _apply_hooks(strategies, hook_dispatcher, context)
     return st.builds(partial(Case, **static_parameters), **strategies)
 
 
-def _apply_hooks(strategies: Dict[str, st.SearchStrategy], getter: Callable[[str], Optional[Hook]]) -> None:
-    for key, strategy in strategies.items():
-        hook = getter(key)
-        if hook is not None:
-            strategies[key] = hook(strategy)
+def _apply_hooks(strategies: Dict[str, st.SearchStrategy], dispatcher: HookDispatcher, context: HookContext) -> None:
+    for key in strategies:
+        for hook in dispatcher.get_all_by_name(f"before_generate_{key}"):
+            # Get the strategy on each hook to pass the first hook output as an input to the next one
+            strategy = strategies[key]
+            args: Union[Tuple[st.SearchStrategy], Tuple[HookContext, st.SearchStrategy]]
+            if _accepts_context(hook):
+                args = (context, strategy)
+            else:
+                args = (strategy,)
+            strategies[key] = hook(*args)
+
+
+def _accepts_context(hook: Hook) -> bool:
+    # There are no restrictions on the first argument's name and we don't check its name here.
+    return len(inspect.signature(hook).parameters) == 2
 
 
 def register_string_format(name: str, strategy: st.SearchStrategy) -> None:
@@ -208,3 +224,9 @@ def register_string_format(name: str, strategy: st.SearchStrategy) -> None:
 def init_default_strategies() -> None:
     register_string_format("binary", st.binary())
     register_string_format("byte", st.binary().map(lambda x: b64encode(x).decode()))
+
+    def make_basic_auth_str(item: Tuple[str, str]) -> str:
+        return _basic_auth_str(*item)
+
+    register_string_format("_basic_auth", st.tuples(st.text(), st.text()).map(make_basic_auth_str))  # type: ignore
+    register_string_format("_bearer_auth", st.text().map("Bearer {}".format))
