@@ -1,7 +1,7 @@
 import os
 import platform
 import shutil
-from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast
+from typing import Any, Dict, List, Optional, Tuple, Union, cast
 
 import click
 from hypothesis import settings
@@ -10,8 +10,9 @@ from ..._compat import metadata
 from ...constants import __version__
 from ...models import Status
 from ...runner import events
-from ...runner.serialization import SerializedCase, SerializedCheck, SerializedTestResult
+from ...runner.serialization import SerializedCase, SerializedTestResult
 from ..context import ExecutionContext
+from ..handlers import EventHandler, get_unique_failures
 
 
 def get_terminal_width() -> int:
@@ -147,7 +148,7 @@ def display_failures(context: ExecutionContext, event: events.Finished) -> None:
 def display_failures_for_single_test(result: SerializedTestResult) -> None:
     """Display a failure for a single method / endpoint."""
     display_subsection(result)
-    checks = _get_unique_failures(result.checks)
+    checks = get_unique_failures(result.checks)
     for idx, check in enumerate(checks, 1):
         message: Optional[str]
         if check.message:
@@ -161,18 +162,6 @@ def display_failures_for_single_test(result: SerializedTestResult) -> None:
             click.echo("\n")
 
 
-def _get_unique_failures(checks: List[SerializedCheck]) -> List[SerializedCheck]:
-    """Return only unique checks that should be displayed in the output."""
-    seen: Set[Tuple[str, Optional[str]]] = set()
-    unique_checks = []
-    for check in reversed(checks):
-        # There are also could be checks that didn't fail
-        if check.example is not None and check.value == Status.failure and (check.name, check.message) not in seen:
-            unique_checks.append(check)
-            seen.add((check.name, check.message))
-    return unique_checks
-
-
 def display_example(
     case: SerializedCase, check_name: Optional[str] = None, message: Optional[str] = None, seed: Optional[int] = None
 ) -> None:
@@ -216,22 +205,36 @@ def display_single_log(result: SerializedTestResult) -> None:
     click.echo("\n\n".join(result.logs))
 
 
-def display_statistic(event: events.Finished) -> None:
+def display_statistic(context: ExecutionContext, event: events.Finished) -> None:
     """Format and print statistic collected by :obj:`models.TestResult`."""
     display_section_name("SUMMARY")
     click.echo()
     total = event.total
     if event.is_empty or not total:
         click.secho("No checks were performed.", bold=True)
-        return
 
+    if total:
+        display_checks_statistics(total)
+
+    if context.cassette_file_name or context.junit_xml_file:
+        click.echo()
+
+    if context.cassette_file_name:
+        category = click.style("Network log", bold=True)
+        click.secho(f"{category}: {context.cassette_file_name}")
+
+    if context.junit_xml_file:
+        category = click.style("JUnit XML file", bold=True)
+        click.secho(f"{category}: {context.junit_xml_file}")
+
+
+def display_checks_statistics(total: Dict[str, Dict[Union[str, Status], int]]) -> None:
     padding = 20
     col1_len = max(map(len, total.keys())) + padding
     col2_len = len(str(max(total.values(), key=lambda v: v["total"])["total"])) * 2 + padding
     col3_len = padding
-
-    template = f"{{:{col1_len}}}{{:{col2_len}}}{{:{col3_len}}}"
-
+    click.secho("Performed checks:", bold=True)
+    template = f"    {{:{col1_len}}}{{:{col2_len}}}{{:{col3_len}}}"
     for check_name, results in total.items():
         display_check_result(check_name, results, template)
 
@@ -246,11 +249,7 @@ def display_check_result(check_name: str, results: Dict[Union[str, Status], int]
         color = "green"
     success = results.get(Status.success, 0)
     total = results.get("total", 0)
-    click.echo(
-        template.format(
-            click.style(check_name, bold=True), f"{success} / {total} passed", click.style(verdict, fg=color, bold=True)
-        )
-    )
+    click.echo(template.format(check_name, f"{success} / {total} passed", click.style(verdict, fg=color, bold=True)))
 
 
 def display_internal_error(context: ExecutionContext, event: events.InternalError) -> None:
@@ -260,7 +259,19 @@ def display_internal_error(context: ExecutionContext, event: events.InternalErro
             message = event.exception_with_traceback
         else:
             message = event.exception
-        click.secho(f"Error: {message}", fg="red")
+        message = (
+            f"Error: {message}\n"
+            f"Add this option to your command line parameters to see full tracebacks: --show-errors-tracebacks"
+        )
+        if event.exception_type == "jsonschema.exceptions.ValidationError":
+            message += (
+                "\n\nYou can disable input schema validation with --validate-schema=false "
+                "command-line option\nIn this case, Schemathesis can not guarantee proper"
+                " behavior during the test run"
+            )
+        click.secho(
+            message, fg="red",
+        )
 
 
 def handle_initialized(context: ExecutionContext, event: events.Initialized) -> None:
@@ -314,7 +325,7 @@ def handle_finished(context: ExecutionContext, event: events.Finished) -> None:
     display_errors(context, event)
     display_failures(context, event)
     display_application_logs(context, event)
-    display_statistic(event)
+    display_statistic(context, event)
     click.echo()
     display_summary(event)
 
@@ -329,18 +340,19 @@ def handle_internal_error(context: ExecutionContext, event: events.InternalError
     raise click.Abort
 
 
-def handle_event(context: ExecutionContext, event: events.ExecutionEvent) -> None:
-    """Choose and execute a proper handler for the given event."""
-    if isinstance(event, events.Initialized):
-        handle_initialized(context, event)
-    if isinstance(event, events.BeforeExecution):
-        handle_before_execution(context, event)
-    if isinstance(event, events.AfterExecution):
-        context.hypothesis_output.extend(event.hypothesis_output)
-        handle_after_execution(context, event)
-    if isinstance(event, events.Finished):
-        handle_finished(context, event)
-    if isinstance(event, events.Interrupted):
-        handle_interrupted(context, event)
-    if isinstance(event, events.InternalError):
-        handle_internal_error(context, event)
+class DefaultOutputStyleHandler(EventHandler):
+    def handle_event(self, context: ExecutionContext, event: events.ExecutionEvent) -> None:
+        """Choose and execute a proper handler for the given event."""
+        if isinstance(event, events.Initialized):
+            handle_initialized(context, event)
+        if isinstance(event, events.BeforeExecution):
+            handle_before_execution(context, event)
+        if isinstance(event, events.AfterExecution):
+            context.hypothesis_output.extend(event.hypothesis_output)
+            handle_after_execution(context, event)
+        if isinstance(event, events.Finished):
+            handle_finished(context, event)
+        if isinstance(event, events.Interrupted):
+            handle_interrupted(context, event)
+        if isinstance(event, events.InternalError):
+            handle_internal_error(context, event)
