@@ -1,9 +1,15 @@
 from copy import deepcopy
 from typing import Any, Dict
 
+from .utils import traverse_schema
+
 
 def to_json_schema(schema: Dict[str, Any], nullable_name: str) -> Dict[str, Any]:
-    """Convert Open API parameters to JSON Schema."""
+    """Convert Open API parameters to JSON Schema.
+
+    NOTE. This function is applied to all keywords (including nested) during schema resolving, thus it is not recursive.
+    See a recursive version below.
+    """
     schema = deepcopy(schema)
     if schema.get(nullable_name) is True:
         del schema[nullable_name]
@@ -18,16 +24,8 @@ def to_json_schema(schema: Dict[str, Any], nullable_name: str) -> Dict[str, Any]
     if schema.get("type") == "file":
         schema["type"] = "string"
         schema["format"] = "binary"
-    _handle_boundaries(schema, "maximum", "exclusiveMaximum")
-    _handle_boundaries(schema, "minimum", "exclusiveMinimum")
     return schema
 
 
-def _handle_boundaries(schema: Dict[str, Any], boundary_name: str, boundary_exclusive_name: str) -> None:
-    # Replace exclusive field only if it is True
-    # if it is non boolean, then leave as is
-    exclusive_maximum = schema.get(boundary_exclusive_name)
-    if exclusive_maximum is True:
-        schema[boundary_exclusive_name] = schema.pop(boundary_name)
-    elif exclusive_maximum is False:
-        del schema[boundary_exclusive_name]
+def to_json_schema_recursive(schema: Dict[str, Any], nullable_name: str) -> Dict[str, Any]:
+    return traverse_schema(schema, to_json_schema, nullable_name)
