@@ -3,6 +3,7 @@ from urllib.parse import urlparse
 
 import hypothesis.errors
 
+from .. import fixups as _fixups
 from .. import loaders
 from ..checks import DEFAULT_CHECKS
 from ..models import CheckFunction
@@ -11,16 +12,20 @@ from ..types import Filter, NotSet, RawAuth
 from ..utils import dict_not_none_values, dict_true_values, file_exists, get_base_url, get_requests_auth, import_app
 from . import events
 from .impl import BaseRunner, SingleThreadRunner, SingleThreadWSGIRunner, ThreadPoolRunner, ThreadPoolWSGIRunner
+from .targeted import DEFAULT_TARGETS, Target
 
 
 def prepare(  # pylint: disable=too-many-arguments
-    schema_uri: str,
+    schema_uri: Union[str, Dict[str, Any]],
     *,
     # Runtime behavior
     checks: Iterable[CheckFunction] = DEFAULT_CHECKS,
+    targets: Iterable[Target] = DEFAULT_TARGETS,
     workers_num: int = 1,
     seed: Optional[int] = None,
     exit_first: bool = False,
+    store_interactions: bool = False,
+    fixups: Iterable[str] = (),
     # Schema loading
     loader: Callable = loaders.from_uri,
     base_url: Optional[str] = None,
@@ -31,6 +36,7 @@ def prepare(  # pylint: disable=too-many-arguments
     endpoint: Optional[Filter] = None,
     method: Optional[Filter] = None,
     tag: Optional[Filter] = None,
+    operation_id: Optional[Filter] = None,
     app: Optional[str] = None,
     validate_schema: bool = True,
     # Hypothesis-specific configuration
@@ -45,6 +51,8 @@ def prepare(  # pylint: disable=too-many-arguments
     """Prepare a generator that will run test cases against the given API definition."""
     # pylint: disable=too-many-locals
 
+    validate_loader(loader, schema_uri)
+
     if auth is None:
         # Auth type doesn't matter if auth is not passed
         auth_type = None  # type: ignore
@@ -64,9 +72,11 @@ def prepare(  # pylint: disable=too-many-arguments
         endpoint=endpoint,
         method=method,
         tag=tag,
+        operation_id=operation_id,
         app=app,
         validate_schema=validate_schema,
         checks=checks,
+        targets=targets,
         hypothesis_options=hypothesis_options,
         seed=seed,
         workers_num=workers_num,
@@ -75,20 +85,43 @@ def prepare(  # pylint: disable=too-many-arguments
         auth_type=auth_type,
         headers=headers,
         request_timeout=request_timeout,
+        store_interactions=store_interactions,
+        fixups=fixups,
     )
 
 
+def validate_loader(loader: Callable, schema_uri: Union[str, Dict[str, Any]]) -> None:
+    """Sanity checking for input schema & loader."""
+    if loader not in (
+        loaders.from_uri,
+        loaders.from_aiohttp,
+        loaders.from_dict,
+        loaders.from_file,
+        loaders.from_path,
+        loaders.from_wsgi,
+    ):
+        # Custom loaders are not checked
+        return
+    if isinstance(schema_uri, dict):
+        if loader is not loaders.from_dict:
+            raise ValueError("Dictionary as a schema is allowed only with `from_dict` loader")
+    elif loader is loaders.from_dict:
+        raise ValueError("Schema should be a dictionary for `from_dict` loader")
+
+
 def execute_from_schema(
     *,
-    schema_uri: str,
+    schema_uri: Union[str, Dict[str, Any]],
     loader: Callable = loaders.from_uri,
     base_url: Optional[str] = None,
     endpoint: Optional[Filter] = None,
     method: Optional[Filter] = None,
     tag: Optional[Filter] = None,
+    operation_id: Optional[Filter] = None,
     app: Optional[str] = None,
     validate_schema: bool = True,
     checks: Iterable[CheckFunction],
+    targets: Iterable[Target],
     workers_num: int = 1,
     hypothesis_options: Dict[str, Any],
     auth: Optional[RawAuth] = None,
@@ -97,6 +130,8 @@ def execute_from_schema(
     request_timeout: Optional[int] = None,
     seed: Optional[int] = None,
     exit_first: bool = False,
+    store_interactions: bool = False,
+    fixups: Iterable[str] = (),
 ) -> Generator[events.ExecutionEvent, None, None]:
     """Execute tests for the given schema.
 
@@ -104,6 +139,12 @@ def execute_from_schema(
     """
     # pylint: disable=too-many-locals
     try:
+        if fixups:
+            if "all" in fixups:
+                _fixups.install()
+            else:
+                _fixups.install(fixups)
+
         if app is not None:
             app = import_app(app)
         schema = load_schema(
@@ -118,6 +159,7 @@ def execute_from_schema(
             endpoint=endpoint,
             method=method,
             tag=tag,
+            operation_id=operation_id,
         )
 
         runner: BaseRunner
@@ -126,6 +168,7 @@ def execute_from_schema(
                 runner = ThreadPoolWSGIRunner(
                     schema=schema,
                     checks=checks,
+                    targets=targets,
                     hypothesis_settings=hypothesis_options,
                     auth=auth,
                     auth_type=auth_type,
@@ -133,11 +176,13 @@ def execute_from_schema(
                     seed=seed,
                     workers_num=workers_num,
                     exit_first=exit_first,
+                    store_interactions=store_interactions,
                 )
             else:
                 runner = ThreadPoolRunner(
                     schema=schema,
                     checks=checks,
+                    targets=targets,
                     hypothesis_settings=hypothesis_options,
                     auth=auth,
                     auth_type=auth_type,
@@ -145,23 +190,27 @@ def execute_from_schema(
                     seed=seed,
                     request_timeout=request_timeout,
                     exit_first=exit_first,
+                    store_interactions=store_interactions,
                 )
         else:
             if schema.app:
                 runner = SingleThreadWSGIRunner(
                     schema=schema,
                     checks=checks,
+                    targets=targets,
                     hypothesis_settings=hypothesis_options,
                     auth=auth,
                     auth_type=auth_type,
                     headers=headers,
                     seed=seed,
                     exit_first=exit_first,
+                    store_interactions=store_interactions,
                 )
             else:
                 runner = SingleThreadRunner(
                     schema=schema,
                     checks=checks,
+                    targets=targets,
                     hypothesis_settings=hypothesis_options,
                     auth=auth,
                     auth_type=auth_type,
@@ -169,6 +218,7 @@ def execute_from_schema(
                     seed=seed,
                     request_timeout=request_timeout,
                     exit_first=exit_first,
+                    store_interactions=store_interactions,
                 )
         yield from runner.execute()
     except Exception as exc:
@@ -176,7 +226,7 @@ def execute_from_schema(
 
 
 def load_schema(
-    schema_uri: str,
+    schema_uri: Union[str, Dict[str, Any]],
     *,
     base_url: Optional[str] = None,
     loader: Callable = loaders.from_uri,
@@ -190,20 +240,24 @@ def load_schema(
     endpoint: Optional[Filter] = None,
     method: Optional[Filter] = None,
     tag: Optional[Filter] = None,
+    operation_id: Optional[Filter] = None,
 ) -> BaseSchema:
     """Load schema via specified loader and parameters."""
-    loader_options = dict_true_values(base_url=base_url, endpoint=endpoint, method=method, tag=tag, app=app)
+    loader_options = dict_true_values(
+        base_url=base_url, endpoint=endpoint, method=method, tag=tag, operation_id=operation_id, app=app
+    )
 
-    if file_exists(schema_uri):
-        loader = loaders.from_path
-    elif app is not None and not urlparse(schema_uri).netloc:
-        # If `schema` is not an existing filesystem path or an URL then it is considered as an endpoint with
-        # the given app
-        loader = loaders.get_loader_for_app(app)
-    else:
-        loader_options.update(dict_true_values(headers=headers, auth=auth, auth_type=auth_type))
+    if not isinstance(schema_uri, dict):
+        if file_exists(schema_uri):
+            loader = loaders.from_path
+        elif app is not None and not urlparse(schema_uri).netloc:
+            # If `schema` is not an existing filesystem path or an URL then it is considered as an endpoint with
+            # the given app
+            loader = loaders.get_loader_for_app(app)
+        else:
+            loader_options.update(dict_true_values(headers=headers, auth=auth, auth_type=auth_type))
 
-    if "base_url" not in loader_options:
+    if "base_url" not in loader_options and not isinstance(schema_uri, dict):
         loader_options["base_url"] = get_base_url(schema_uri)
     if loader is loaders.from_uri and loader_options.get("auth"):
         loader_options["auth"] = get_requests_auth(loader_options["auth"], loader_options.pop("auth_type", None))
