@@ -16,6 +16,7 @@ from ...runner import events
 from ...schemas import BaseSchema
 from ...types import RawAuth
 from ...utils import GenericResponse, capture_hypothesis_output
+from ..targeted import Target
 
 DEFAULT_DEADLINE = 500  # pragma: no mutate
 
@@ -31,11 +32,13 @@ def get_hypothesis_settings(hypothesis_options: Dict[str, Any]) -> hypothesis.se
 class BaseRunner:
     schema: BaseSchema = attr.ib()  # pragma: no mutate
     checks: Iterable[CheckFunction] = attr.ib()  # pragma: no mutate
+    targets: Iterable[Target] = attr.ib()  # pragma: no mutate
     hypothesis_settings: hypothesis.settings = attr.ib(converter=get_hypothesis_settings)  # pragma: no mutate
     auth: Optional[RawAuth] = attr.ib(default=None)  # pragma: no mutate
     auth_type: Optional[str] = attr.ib(default=None)  # pragma: no mutate
     headers: Optional[Dict[str, Any]] = attr.ib(default=None)  # pragma: no mutate
     request_timeout: Optional[int] = attr.ib(default=None)  # pragma: no mutate
+    store_interactions: bool = attr.ib(default=False)  # pragma: no mutate
     seed: Optional[int] = attr.ib(default=None)  # pragma: no mutate
     exit_first: bool = attr.ib(default=False)  # pragma: no mutate
 
@@ -62,25 +65,27 @@ class BaseRunner:
 
 
 def run_test(
-    schema: BaseSchema,
     endpoint: Endpoint,
     test: Union[Callable, InvalidSchema],
     checks: Iterable[CheckFunction],
+    targets: Iterable[Target],
     results: TestResultSet,
+    headers: Optional[Dict[str, Any]],
     **kwargs: Any,
 ) -> Generator[events.ExecutionEvent, None, None]:
     """A single test run with all error handling needed."""
     # pylint: disable=too-many-arguments
-    result = TestResult(endpoint=endpoint)
+    result = TestResult(endpoint=endpoint, overridden_headers=headers)
     yield events.BeforeExecution.from_endpoint(endpoint=endpoint)
     hypothesis_output: List[str] = []
+    test_start_time = time.monotonic()
     try:
         if isinstance(test, InvalidSchema):
             status = Status.error
             result.add_error(test)
         else:
             with capture_hypothesis_output() as hypothesis_output:
-                test(checks, result, **kwargs)
+                test(checks, targets, result, headers=headers, **kwargs)
             status = Status.success
     except (AssertionError, hypothesis.errors.MultipleFailures):
         status = Status.failure
@@ -109,12 +114,15 @@ def run_test(
     except Exception as error:
         status = Status.error
         result.add_error(error)
+    test_elapsed_time = time.monotonic() - test_start_time
     # Fetch seed value, hypothesis generates it during test execution
     result.seed = getattr(test, "_hypothesis_internal_use_seed", None) or getattr(
         test, "_hypothesis_internal_use_generated_seed", None
     )
     results.append(result)
-    yield events.AfterExecution.from_result(result=result, status=status, hypothesis_output=hypothesis_output)
+    yield events.AfterExecution.from_result(
+        result=result, status=status, elapsed_time=test_elapsed_time, hypothesis_output=hypothesis_output
+    )
 
 
 def run_checks(case: Case, checks: Iterable[CheckFunction], result: TestResult, response: GenericResponse) -> None:
@@ -133,30 +141,39 @@ def run_checks(case: Case, checks: Iterable[CheckFunction], result: TestResult,
         raise get_grouped_exception(*errors)
 
 
+def run_targets(targets: Iterable[Target], elapsed: float) -> None:
+    for target in targets:
+        if target == Target.response_time:
+            hypothesis.target(elapsed, label="response_time")
+
+
 def network_test(
     case: Case,
     checks: Iterable[CheckFunction],
+    targets: Iterable[Target],
     result: TestResult,
     session: requests.Session,
     request_timeout: Optional[int],
+    store_interactions: bool,
+    headers: Optional[Dict[str, Any]],
 ) -> None:
     """A single test body that will be executed against the target."""
     # pylint: disable=too-many-arguments
+    headers = headers or {}
+    headers.setdefault("User-Agent", USER_AGENT)
     timeout = prepare_timeout(request_timeout)
-    response = case.call(session=session, timeout=timeout)
+    response = case.call(session=session, headers=headers, timeout=timeout)
+    run_targets(targets, response.elapsed.total_seconds())
+    if store_interactions:
+        result.store_requests_response(response)
     run_checks(case, checks, result, response)
 
 
 @contextmanager
-def get_session(
-    auth: Optional[Union[HTTPDigestAuth, RawAuth]] = None, headers: Optional[Dict[str, Any]] = None
-) -> Generator[requests.Session, None, None]:
+def get_session(auth: Optional[Union[HTTPDigestAuth, RawAuth]] = None) -> Generator[requests.Session, None, None]:
     with requests.Session() as session:
         if auth is not None:
             session.auth = auth
-        session.headers["User-agent"] = USER_AGENT
-        if headers is not None:
-            session.headers.update(**headers)
         yield session
 
 
@@ -171,15 +188,22 @@ def prepare_timeout(timeout: Optional[int]) -> Optional[float]:
 def wsgi_test(
     case: Case,
     checks: Iterable[CheckFunction],
+    targets: Iterable[Target],
     result: TestResult,
     auth: Optional[RawAuth],
     auth_type: Optional[str],
     headers: Optional[Dict[str, Any]],
+    store_interactions: bool,
 ) -> None:
     # pylint: disable=too-many-arguments
     headers = _prepare_wsgi_headers(headers, auth, auth_type)
     with catching_logs(LogCaptureHandler(), level=logging.DEBUG) as recorded:
+        start = time.monotonic()
         response = case.call_wsgi(headers=headers)
+        elapsed = time.monotonic() - start
+    run_targets(targets, elapsed)
+    if store_interactions:
+        result.store_wsgi_response(case, response, headers, elapsed)
     result.logs.extend(recorded.records)
     run_checks(case, checks, result, response)
 
@@ -188,7 +212,7 @@ def _prepare_wsgi_headers(
     headers: Optional[Dict[str, Any]], auth: Optional[RawAuth], auth_type: Optional[str]
 ) -> Dict[str, Any]:
     headers = headers or {}
-    headers.setdefault("User-agent", USER_AGENT)
+    headers.setdefault("User-Agent", USER_AGENT)
     wsgi_auth = get_wsgi_auth(auth, auth_type)
     if wsgi_auth:
         headers["Authorization"] = wsgi_auth
