@@ -22,13 +22,13 @@ import yaml
 from requests.structures import CaseInsensitiveDict
 
 from ._hypothesis import make_test_or_exception
-from .constants import HookLocation
-from .converter import to_json_schema
+from .converter import to_json_schema, to_json_schema_recursive
 from .exceptions import InvalidSchema
-from .filters import should_skip_by_tag, should_skip_endpoint, should_skip_method
-from .models import Endpoint, empty_object
-from .types import Filter, Hook, NotSet
-from .utils import NOT_SET, GenericResponse, StringDatesYAMLLoader
+from .filters import should_skip_by_operation_id, should_skip_by_tag, should_skip_endpoint, should_skip_method
+from .hooks import HookContext, HookDispatcher, HookLocation, HookScope, dispatch, warn_deprecated_hook
+from .models import Endpoint, EndpointDefinition, empty_object
+from .types import Filter, GenericTest, Hook, NotSet
+from .utils import NOT_SET, GenericResponse, StringDatesYAMLLoader, deprecated, traverse_schema
 
 # Reference resolving will stop after this depth
 RECURSION_DEPTH_LIMIT = 100
@@ -52,16 +52,36 @@ def load_file_uri(location: str) -> Dict[str, Any]:
     return load_file_impl(location, urlopen)
 
 
+class ConvertingResolver(jsonschema.RefResolver):
+    """A custom resolver converts resolved OpenAPI schemas to JSON Schema.
+
+    When recursive schemas are validated we need to have resolved documents properly converted.
+    This approach is the simplest one, since this logic isolated in a single place.
+    """
+
+    def __init__(self, *args: Any, conversion_kwargs: Any, **kwargs: Any) -> None:
+        super().__init__(*args, **kwargs)
+        self.conversion_kwargs = conversion_kwargs
+
+    def resolve(self, ref: str) -> Tuple[str, Any]:
+        url, document = super().resolve(ref)
+        document = traverse_schema(document, to_json_schema, **self.conversion_kwargs)
+        return url, document
+
+
 @attr.s()  # pragma: no mutate
 class BaseSchema(Mapping):
+    nullable_name: str
     raw_schema: Dict[str, Any] = attr.ib()  # pragma: no mutate
     location: Optional[str] = attr.ib(default=None)  # pragma: no mutate
     base_url: Optional[str] = attr.ib(default=None)  # pragma: no mutate
     method: Optional[Filter] = attr.ib(default=None)  # pragma: no mutate
     endpoint: Optional[Filter] = attr.ib(default=None)  # pragma: no mutate
     tag: Optional[Filter] = attr.ib(default=None)  # pragma: no mutate
+    operation_id: Optional[Filter] = attr.ib(default=None)  # pragma: no mutate
     app: Any = attr.ib(default=None)  # pragma: no mutate
-    hooks: Dict[HookLocation, Hook] = attr.ib(factory=dict)  # pragma: no mutate
+    hooks: HookDispatcher = attr.ib(factory=lambda: HookDispatcher(scope=HookScope.SCHEMA))  # pragma: no mutate
+    test_function: Optional[GenericTest] = attr.ib(default=None)  # pragma: no mutate
     validate_schema: bool = attr.ib(default=True)  # pragma: no mutate
 
     def __iter__(self) -> Iterator[str]:
@@ -93,8 +113,11 @@ class BaseSchema(Mapping):
     def resolver(self) -> jsonschema.RefResolver:
         if not hasattr(self, "_resolver"):
             # pylint: disable=attribute-defined-outside-init
-            self._resolver = jsonschema.RefResolver(
-                self.location or "", self.raw_schema, handlers={"file": load_file_uri, "": load_file}
+            self._resolver = ConvertingResolver(
+                self.location or "",
+                self.raw_schema,
+                conversion_kwargs={"nullable_name": self.nullable_name},
+                handlers={"file": load_file_uri, "": load_file},
             )
         return self._resolver
 
@@ -114,26 +137,30 @@ class BaseSchema(Mapping):
             test = make_test_or_exception(endpoint, func, settings, seed)
             yield endpoint, test
 
-    def parametrize(
+    def parametrize(  # pylint: disable=too-many-arguments
         self,
         method: Optional[Filter] = NOT_SET,
         endpoint: Optional[Filter] = NOT_SET,
         tag: Optional[Filter] = NOT_SET,
+        operation_id: Optional[Filter] = NOT_SET,
         validate_schema: Union[bool, NotSet] = NOT_SET,
     ) -> Callable:
         """Mark a test function as a parametrized one."""
 
-        def wrapper(func: Callable) -> Callable:
-            func._schemathesis_test = self.clone(method, endpoint, tag, validate_schema)  # type: ignore
+        def wrapper(func: GenericTest) -> GenericTest:
+            HookDispatcher.add_dispatcher(func)
+            func._schemathesis_test = self.clone(func, method, endpoint, tag, operation_id, validate_schema)  # type: ignore
             return func
 
         return wrapper
 
-    def clone(
+    def clone(  # pylint: disable=too-many-arguments
         self,
+        test_function: Optional[GenericTest] = None,
         method: Optional[Filter] = NOT_SET,
         endpoint: Optional[Filter] = NOT_SET,
         tag: Optional[Filter] = NOT_SET,
+        operation_id: Optional[Filter] = NOT_SET,
         validate_schema: Union[bool, NotSet] = NOT_SET,
     ) -> "BaseSchema":
         if method is NOT_SET:
@@ -142,6 +169,8 @@ class BaseSchema(Mapping):
             endpoint = self.endpoint
         if tag is NOT_SET:
             tag = self.tag
+        if operation_id is NOT_SET:
+            operation_id = self.operation_id
         if validate_schema is NOT_SET:
             validate_schema = self.validate_schema
 
@@ -152,30 +181,58 @@ class BaseSchema(Mapping):
             method=method,
             endpoint=endpoint,
             tag=tag,
+            operation_id=operation_id,
             app=self.app,
             hooks=self.hooks,
+            test_function=test_function,
             validate_schema=validate_schema,  # type: ignore
         )
 
-    def _get_response_schema(self, definition: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+    def _get_response_schema(
+        self, definition: Dict[str, Any], scope: str
+    ) -> Tuple[List[str], Optional[Dict[str, Any]]]:
         """Extract response schema from `responses`."""
         raise NotImplementedError
 
+    @deprecated("'register_hook` is deprecated, use `hooks.register' instead")
     def register_hook(self, place: str, hook: Hook) -> None:
-        key = HookLocation[place]
-        self.hooks[key] = hook
+        warn_deprecated_hook(hook)
+        if place not in HookLocation.__members__:
+            raise KeyError(place)
+        self.hooks.register_hook_with_name(f"before_generate_{place}", hook, skip_validation=True)
 
-    def get_hook(self, place: str) -> Optional[Hook]:
-        key = HookLocation[place]
-        return self.hooks.get(key)
+    @deprecated("'with_hook` is deprecated, use `hooks.apply' instead")
+    def with_hook(self, place: str, hook: Hook) -> Callable[[GenericTest], GenericTest]:
+        """Register a hook for a specific test."""
+        warn_deprecated_hook(hook)
+        if place not in HookLocation.__members__:
+            raise KeyError(place)
+
+        return self.hooks.apply(f"before_generate_{place}", hook, skip_validation=True)
+
+    def get_local_hook_dispatcher(self) -> Optional[HookDispatcher]:
+        """Get a HookDispatcher instance bound to the test if present."""
+        # It might be not present when it is used without pytest via `Endpoint.as_strategy()`
+        if self.test_function is not None:
+            return self.test_function._schemathesis_hooks  # type: ignore
+        return None
+
+    def dispatch_hook(self, name: str, context: HookContext, *args: Any, **kwargs: Any) -> None:
+        """Dispatch a hook via all available dispatchers."""
+        dispatch(name, context, *args, **kwargs)
+        self.hooks.dispatch(name, context, *args, **kwargs)
+        local_dispatcher = self.get_local_hook_dispatcher()
+        if local_dispatcher is not None:
+            local_dispatcher.dispatch(name, context, *args, **kwargs)
 
     def get_content_types(self, endpoint: Endpoint, response: GenericResponse) -> List[str]:
         """Content types available for this endpoint."""
         raise NotImplementedError
 
 
-class SwaggerV20(BaseSchema):
+class SwaggerV20(BaseSchema):  # pylint: disable=too-many-public-methods
     nullable_name = "x-nullable"
+    example_field = "x-example"
     operations: Tuple[str, ...] = ("get", "put", "post", "delete", "options", "head", "patch")
 
     def __repr__(self) -> str:
@@ -205,39 +262,114 @@ class SwaggerV20(BaseSchema):
     def get_all_endpoints(self) -> Generator[Endpoint, None, None]:
         try:
             paths = self.raw_schema["paths"]  # pylint: disable=unsubscriptable-object
+            context = HookContext()
             for path, methods in paths.items():
                 full_path = self.get_full_path(path)
                 if should_skip_endpoint(full_path, self.endpoint):
                     continue
+                self.dispatch_hook("before_process_path", context, path, methods)
+                # We need to know a proper scope in what methods are.
+                # It will allow us to provide a proper reference resolving in `response_schema_conformance` and avoid
+                # recursion errors
+                if "$ref" in methods:
+                    scope, raw_methods = deepcopy(self.resolver.resolve(methods["$ref"]))
+                else:
+                    raw_methods, scope = deepcopy(methods), self.resolver.resolution_scope
                 methods = self.resolve(methods)
                 common_parameters = get_common_parameters(methods)
-                for method, definition in methods.items():
+                for method, resolved_definition in methods.items():
                     # Only method definitions are parsed
                     if (
                         method not in self.operations
                         or should_skip_method(method, self.method)
-                        or should_skip_by_tag(definition.get("tags"), self.tag)
+                        or should_skip_by_tag(resolved_definition.get("tags"), self.tag)
+                        or should_skip_by_operation_id(resolved_definition.get("operationId"), self.operation_id)
                     ):
                         continue
-                    parameters = itertools.chain(definition.get("parameters", ()), common_parameters)
-                    yield self.make_endpoint(full_path, method, parameters, definition)
+                    parameters = itertools.chain(resolved_definition.get("parameters", ()), common_parameters)
+                    # To prevent recursion errors we need to pass not resolved schema as well
+                    # It could be used for response validation
+                    raw_definition = EndpointDefinition(raw_methods[method], scope)
+                    yield self.make_endpoint(full_path, method, parameters, resolved_definition, raw_definition)
         except (KeyError, AttributeError, jsonschema.exceptions.RefResolutionError):
             raise InvalidSchema("Schema parsing failed. Please check your schema.")
 
-    def make_endpoint(
-        self, full_path: str, method: str, parameters: Iterator[Dict[str, Any]], definition: Dict[str, Any]
+    def make_endpoint(  # pylint: disable=too-many-arguments
+        self,
+        full_path: str,
+        method: str,
+        parameters: Iterator[Dict[str, Any]],
+        resolved_definition: Dict[str, Any],
+        raw_definition: EndpointDefinition,
     ) -> Endpoint:
         """Create JSON schemas for query, body, etc from Swagger parameters definitions."""
         base_url = self.base_url
         if base_url is not None:
             base_url = base_url.rstrip("/")  # pragma: no mutate
         endpoint = Endpoint(
-            path=full_path, method=method.upper(), definition=definition, base_url=base_url, app=self.app, schema=self
+            path=full_path,
+            method=method.upper(),
+            definition=raw_definition,
+            base_url=base_url,
+            app=self.app,
+            schema=self,
         )
         for parameter in parameters:
             self.process_parameter(endpoint, parameter)
+        self.process_security_definitions(endpoint)
         return endpoint
 
+    def get_security_definitions(self) -> Dict[str, Any]:
+        """Extract security definitions from the schema."""
+        return self.raw_schema.get("securityDefinitions", {})
+
+    def get_security_requirements(self, endpoint: Endpoint) -> List[str]:
+        # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/2.0.md#operation-object
+        # > This definition overrides any declared top-level security.
+        # > To remove a top-level security declaration, an empty array can be used.
+        global_requirements = self.raw_schema.get("security", [])
+        local_requirements = endpoint.definition.raw.get("security", None)
+        if local_requirements is not None:
+            requirements = local_requirements
+        else:
+            requirements = global_requirements
+        return [key for requirement in requirements for key in requirement]
+
+    def process_security_definitions(self, endpoint: Endpoint) -> None:
+        """Add relevant security parameters to data generation."""
+        definitions = self.get_security_definitions()
+        requirements = self.get_security_requirements(endpoint)
+        for name, definition in definitions.items():
+            if name in requirements:
+                if definition["type"] == "apiKey":
+                    self.process_api_key_security_definition(definition, endpoint)
+                self.process_http_security_definition(definition, endpoint)
+
+    def process_api_key_security_definition(self, definition: Dict[str, Any], endpoint: Endpoint) -> None:
+        if definition["in"] == "query":
+            endpoint.query = self.add_security_definition(endpoint.query, definition)
+        elif definition["in"] == "header":
+            endpoint.headers = self.add_security_definition(endpoint.headers, definition)
+
+    def process_http_security_definition(self, definition: Dict[str, Any], endpoint: Endpoint) -> None:
+        if definition["type"] == "basic":
+            endpoint.headers = self.add_http_auth_definition(endpoint.headers)
+
+    def add_security_definition(
+        self, container: Optional[Dict[str, Any]], definition: Dict[str, Any]
+    ) -> Dict[str, Any]:
+        name = definition["name"]
+        container = container or empty_object()
+        container["properties"][name] = {"name": name, "type": "string"}
+        container["required"].append(name)
+        return container
+
+    def add_http_auth_definition(self, container: Optional[Dict[str, Any]], scheme: str = "basic") -> Dict[str, Any]:
+        container = container or empty_object()
+        container["properties"]["Authorization"] = {"type": "string", "format": f"_{scheme}_auth"}
+        container["required"].append("Authorization")
+        return container
+
     def process_parameter(self, endpoint: Endpoint, parameter: Dict[str, Any]) -> None:
         """Convert each Parameter object to a JSON schema."""
         parameter = deepcopy(parameter)
@@ -280,6 +412,12 @@ class SwaggerV20(BaseSchema):
         container["properties"][name] = self.parameter_to_json_schema(parameter)
         if parameter.get("required", False):
             container["required"].append(name)
+        return self.add_examples(container, parameter)
+
+    def add_examples(self, container: Dict[str, Any], parameter: Dict[str, Any]) -> Dict[str, Any]:
+        if self.example_field in parameter:
+            examples = container.setdefault("example", {})  # examples should be merged together
+            examples[parameter["name"]] = parameter[self.example_field]
         return container
 
     def parameter_to_json_schema(self, data: Dict[str, Any]) -> Dict[str, Any]:
@@ -318,15 +456,33 @@ class SwaggerV20(BaseSchema):
                 item[idx] = self.resolve(sub_item, recursion_level)
         return item
 
+    def resolve_in_scope(self, definition: Dict[str, Any], scope: str) -> Tuple[List[str], Dict[str, Any]]:
+        scopes = [scope]
+        # if there is `$ref` then we have a scope change that should be used during validation later to
+        # resolve nested references correctly
+        if "$ref" in definition:
+            with self.resolver.in_scope(scope):
+                new_scope, definition = deepcopy(self.resolver.resolve(definition["$ref"]))
+            scopes.append(new_scope)
+        return scopes, definition
+
     def prepare(self, item: Dict[str, Any]) -> Dict[str, Any]:
         """Parse schema extension, e.g. "x-nullable" field."""
         return to_json_schema(item, self.nullable_name)
 
-    def _get_response_schema(self, definition: Dict[str, Any]) -> Optional[Dict[str, Any]]:
-        return definition.get("schema")
+    def _get_response_schema(
+        self, definition: Dict[str, Any], scope: str
+    ) -> Tuple[List[str], Optional[Dict[str, Any]]]:
+        scopes, definition = self.resolve_in_scope(deepcopy(definition), scope)
+        schema = definition.get("schema")
+        if not schema:
+            return scopes, None
+        # Extra conversion to JSON Schema is needed here if there was one $ref in the input
+        # because it is not converted
+        return scopes, to_json_schema_recursive(schema, self.nullable_name)
 
     def get_content_types(self, endpoint: Endpoint, response: GenericResponse) -> List[str]:
-        produces = endpoint.definition.get("produces", None)
+        produces = endpoint.definition.raw.get("produces", None)
         if produces:
             return produces
         return self.raw_schema.get("produces", [])
@@ -334,6 +490,7 @@ class SwaggerV20(BaseSchema):
 
 class OpenApi30(SwaggerV20):  # pylint: disable=too-many-ancestors
     nullable_name = "nullable"
+    example_field = "example"
     operations = SwaggerV20.operations + ("trace",)
 
     @property
@@ -359,33 +516,48 @@ class OpenApi30(SwaggerV20):  # pylint: disable=too-many-ancestors
             path += "/"
         return path
 
-    def make_endpoint(
-        self, full_path: str, method: str, parameters: Iterator[Dict[str, Any]], definition: Dict[str, Any]
+    def make_endpoint(  # pylint: disable=too-many-arguments
+        self,
+        full_path: str,
+        method: str,
+        parameters: Iterator[Dict[str, Any]],
+        resolved_definition: Dict[str, Any],
+        raw_definition: EndpointDefinition,
     ) -> Endpoint:
         """Create JSON schemas for query, body, etc from Swagger parameters definitions."""
-        endpoint = super().make_endpoint(full_path, method, parameters, definition)
-        if "requestBody" in definition:
-            self.process_body(endpoint, definition["requestBody"])
+        endpoint = super().make_endpoint(full_path, method, parameters, resolved_definition, raw_definition)
+        if "requestBody" in resolved_definition:
+            self.process_body(endpoint, resolved_definition["requestBody"])
         return endpoint
 
+    def get_security_definitions(self) -> Dict[str, Any]:
+        components = self.raw_schema.get("components", {})
+        security_schemes = components.get("securitySchemes", {})
+        return self.resolve(security_schemes, RECURSION_DEPTH_LIMIT)
+
+    def process_api_key_security_definition(self, definition: Dict[str, Any], endpoint: Endpoint) -> None:
+        if definition["in"] == "cookie":
+            endpoint.cookies = self.add_security_definition(endpoint.cookies, definition)
+        super().process_api_key_security_definition(definition, endpoint)
+
+    def process_http_security_definition(self, definition: Dict[str, Any], endpoint: Endpoint) -> None:
+        if definition["type"] == "http":
+            endpoint.headers = self.add_http_auth_definition(endpoint.headers, scheme=definition["scheme"].lower())
+
     def process_by_type(self, endpoint: Endpoint, parameter: Dict[str, Any]) -> None:
         if parameter["in"] == "cookie":
             self.process_cookie(endpoint, parameter)
         else:
             super().process_by_type(endpoint, parameter)
 
-    def add_parameter(self, container: Optional[Dict[str, Any]], parameter: Dict[str, Any]) -> Dict[str, Any]:
-        container = super().add_parameter(container, parameter)
-        if "example" in parameter["schema"]:
+    def add_examples(self, container: Dict[str, Any], parameter: Dict[str, Any]) -> Dict[str, Any]:
+        if self.example_field in parameter["schema"]:
             examples = container.setdefault("example", {})  # examples should be merged together
-            examples[parameter["name"]] = parameter["schema"]["example"]
+            examples[parameter["name"]] = parameter["schema"][self.example_field]
         # https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.2.md#parameter-object
         # > Furthermore, if referencing a schema which contains an example,
         # > the example value SHALL override the example provided by the schema
-        if "example" in parameter:
-            examples = container.setdefault("example", {})  # examples should be merged together
-            examples[parameter["name"]] = parameter["example"]
-        return container
+        return super().add_examples(container, parameter)
 
     def process_cookie(self, endpoint: Endpoint, parameter: Dict[str, Any]) -> None:
         endpoint.cookies = self.add_parameter(endpoint.cookies, parameter)
@@ -405,16 +577,21 @@ class OpenApi30(SwaggerV20):  # pylint: disable=too-many-ancestors
         # "schema" field is required for all parameters in Open API 3.0
         return super().parameter_to_json_schema(data["schema"])
 
-    def _get_response_schema(self, definition: Dict[str, Any]) -> Optional[Dict[str, Any]]:
+    def _get_response_schema(
+        self, definition: Dict[str, Any], scope: str
+    ) -> Tuple[List[str], Optional[Dict[str, Any]]]:
+        scopes, definition = self.resolve_in_scope(deepcopy(definition), scope)
         options = iter(definition.get("content", {}).values())
         option = next(options, None)
         if option:
-            return option["schema"]
-        return None
+            # Extra conversion to JSON Schema is needed here if there was one $ref in the input
+            # because it is not converted
+            return scopes, to_json_schema_recursive(option["schema"], self.nullable_name)
+        return scopes, None
 
     def get_content_types(self, endpoint: Endpoint, response: GenericResponse) -> List[str]:
         try:
-            responses = endpoint.definition["responses"]
+            responses = endpoint.definition.raw["responses"]
         except KeyError:
             # Possible to get if `validate_schema=False` is passed during schema creation
             raise InvalidSchema("Schema parsing failed. Please check your schema.")
